/* ___  __    ___  ___  ___ 
  / _ \/ _\  / _ \/   \/ __\
 / /_)/\ \  / /_)/ /\ / /         Version 1.0 by mappy
/ ___/ _\ \/ ___/ /_// /___       Compile with MinPSPW 0.11 or newer
\/     \__/\/  /___,'\___*/ 

#include <pspkernel.h>
#include <oslib/oslib.h>
#include <string>
#include <vector>
#include <map>
#include "nmdc.h"
#include "mUtil.h"
using namespace std;

PSP_MODULE_INFO("PSPDC", 0, 1, 0);
PSP_MAIN_THREAD_ATTR(THREAD_ATTR_USER | THREAD_ATTR_VFPU);
PSP_HEAP_SIZE_KB(6*1024);

//////////////////////////////////
// Globals
//////////////////////////////////

#define AREASZ_W     75
#define AREASZ_H     30
#define INPUT_MAXLEN 512
#define USERLIST_X   392
#define USERLIST_XMX 14

enum pspdc_states {
	STATE_FATALWARNING,
	STATE_WIFI,
	STATE_NICK,
	STATE_ADDRESS,
	STATE_PASSWORD,
	STATE_TEXT	
};

static vector<string> lnbuff;
static string inbuff;

static int runningFlag = 1;
static OSL_IMAGE *bkg = NULL, *usr = NULL;

static NMDCHub* h = 0;

#define SAYF(fmt, ...) say(format("%s " fmt, timestamp().c_str(), __VA_ARGS__ ).c_str())
#define SAY(str) say(format("%s " str, timestamp().c_str()).c_str())

//////////////////////////////////
// PSP Framework
//////////////////////////////////

// Exit callback
int exit_callback(int, int, void*) {
    runningFlag = 0;
    return 0;
}

// Callback thread
int CallbackThread(SceSize, void*) {
    int cbid;
    cbid = sceKernelCreateCallback("Exit Callback", exit_callback, NULL);
    sceKernelRegisterExitCallback(cbid);
    sceKernelSleepThreadCB();
    return 0;
}

// Sets up the callback thread and returns its thread id
int SetupCallbacks() {
    int thid = 0;
    thid = sceKernelCreateThread(
		"update_thread", CallbackThread, 0x11, 0xFA0, PSP_THREAD_ATTR_USER, 0
	);
    if (thid >= 0) sceKernelStartThread(thid, 0, 0);
    return thid;
}

// Initialise OSLib
void initOSLib() {
    oslInit(0);
    oslInitGfx(OSL_PF_8888, 1);
    oslInitAudio();
    oslSetQuitOnLoadFailure(1);
    oslSetKeyAutorepeatInit(40);
    oslSetKeyAutorepeatInterval(10);
}

//////////////////////////////////
// Helpers
//////////////////////////////////

void say(const char* msg) {
	string sm(msg);
	long slen = sm.length();
	size_t p = sm.find('\n');
	
	if (p != string::npos) { // Remove newlines
		say(sm.substr(0, p).c_str());
		say(sm.substr(p+1).c_str());
	} else if (slen > AREASZ_W) { // Split to multiple lines
		say(sm.substr(0, AREASZ_W).c_str());
		say(sm.substr(AREASZ_W).c_str());
	} else { // Shuffle buffer
		for (int i = 1; i < AREASZ_H; i++) {
			lnbuff[i-1] = lnbuff[i];
		}
		lnbuff[AREASZ_H-1] = msg;
	}
}

void renderScreen() {
	oslStartDrawing();
	oslDrawImageXY(bkg, 0, 0);
	
	// Draw text
	for (int i = 0; i < AREASZ_H; i++) {
		oslDrawString(8, (8*i)+8, lnbuff[i].c_str());
	}
	
	// Draw userlist
	/*
	int c =0;
	oslDrawImageXY(usr, USERLIST_X, (8*c)+8 );
	oslDrawString(USERLIST_X + 8, (8*c)+8, "example");
	*/
	
	if (h) {
		map<string,NMDCUser>* users = h->GetUsers(); int c = 0;
		if (users->size()) {
			for (
				map<string,NMDCUser>::iterator i = users->begin(), e = users->end();
				i != e; i++, c++
			) {
				if (c > AREASZ_H) break;
				oslDrawImageXY(usr, USERLIST_X, (8*c)+8 );
				if (i->first.length() > USERLIST_XMX) {
					oslDrawString(USERLIST_X + 8, (8*c)+8, i->first.substr(0,USERLIST_XMX).c_str());
				} else {
					oslDrawString(USERLIST_X + 8, (8*c)+8, i->first.c_str());
				}
			}
		}
	}
		
	// Draw input buffer
	oslDrawString(8, (AREASZ_H+2)*8, inbuff.c_str());

	// Draw overlays	
	if (oslOskIsActive()) {
		oslDrawOsk();
	}
	if (oslGetDialogType()) {
		oslDrawDialog();
	}
	
	oslEndDrawing();
}

// FIXME: also clear userlist
#define RESTORE_STATE { \
	delete h; \
	h = NULL; \
	SAY("** Please enter a nick to connect as."); \
	inbuff = nick; \
	state = STATE_NICK; \
}

//////////////////////////////////
// Entry point:
//////////////////////////////////

int main() {
	int frameskip = 0;
	enum pspdc_states state = STATE_WIFI;
	string nick = "", address = "";
	bool debug = false;

	// INITIALISATION
	//////////////////
	
	// Framework
    SetupCallbacks();
    initOSLib();
    oslIntraFontInit(INTRAFONT_CACHE_MED);
    oslNetInit();

    // Load background image
    bkg = oslLoadImageFileJPG((char*)"bg.jpg",   OSL_IN_RAM | OSL_SWIZZLED, OSL_PF_8888);
	usr = oslLoadImageFileJPG((char*)"user.jpg", OSL_IN_RAM | OSL_SWIZZLED, OSL_PF_8888);

    // Load font
    OSL_FONT *font = oslLoadFontFile("tiny.oft");
    oslSetFont(font);
	oslSetTextColor(RGBA(0,0,0,255));
	oslSetBkColor(RGBA(255,255,255,255));

	// Clear text buffers
	inbuff.clear();
	lnbuff.clear();
	for (int i = 0; i < AREASZ_H; i++) lnbuff.push_back("");
	
	// Default messages
	SAY("** Welcome to PSPDC!");
		
	// SELECT WIFI
	//////////////////
	
	SAY("** Connecting to WiFi network..");
    if (!oslIsWlanPowerOn()) {
		SAY("** WARNING: WiFi switch is currently turned off.");
	}

    // Get connections list:
    struct oslNetConfig configs[OSL_MAX_NET_CONFIGS];
    int numconfigs = oslGetNetConfigs(configs);
    if (!numconfigs){
		oslInitMessageDialog("Couldn't find any WiFi configurations!", 0);
		state = STATE_FATALWARNING;
    } else {
		// Initialise WiFi dialog
		oslInitNetDialog();
	}
	
    while(runningFlag && !osl_quit){
		
		// UPDATE SCREEN
		//////////////////
	
        if (!frameskip) {
			renderScreen();
		}
        oslEndFrame();
        frameskip = oslSyncFrame();

		// PULSE NETWORK
		//////////////////
		
		if (h) {
			if (! h->Connected()) {
				SAY("** Network connection closed.");
				RESTORE_STATE;
				
			} else {		
				NMDCEvent ev = h->NextEvent();
				if (ev.type != NMDC_NONE) {
					switch (ev.type) {
						case NMDC_PUBLIC_CHAT: {
							SAYF("<%s> %s", ev.data1.c_str(), ev.data2.c_str());
						} break;
						
						case NMDC_PRIVATE_MSG: {
							SAYF("[%s -> %s] %s", ev.data1.c_str(), nick.c_str(), ev.data2.c_str());
						} break;
						
						case NMDC_SYSTEM_MSG: {
							SAYF("* %s", ev.data2.c_str());
						} break;
						
						case NMDC_USER_JOIN: {
							SAYF("* Joins: %s", ev.data1.c_str());
							// add to userlist
						} break;
						
						case NMDC_USER_PART: {
							SAYF("* Parts: %s", ev.data1.c_str());
							// remove from userlist
						} break;
						
						case NMDC_USER_UPDATE: {
							// Don't care
						} break;
						
						case NMDC_NEED_PASS: {
							SAY("** This username is passworded. Please enter a password to continue.");
							state = STATE_PASSWORD;
						} break;
					
						case NMDC_KICKED: {
							SAY("** The connection was closed by the remote server.");
							RESTORE_STATE;
						} break;
					
						case NMDC_DEBUG_INFO:					
						default: {
							if (debug) {
								SAYF("EVENT %d: [%s][%s]", ev.type, ev.data1.c_str(), ev.data2.c_str() );
							}
						} break;
					}
					
				}
				
			}
		}
		
		// HANDLE INPUT
		//////////////////
		
        oslReadKeys();
		
		// Check if there's an OSK open		
		if (oslOskIsActive()) {
			if (oslGetOskStatus() == PSP_UTILITY_DIALOG_NONE) {
				// OSK finished
				if (oslOskGetResult() != OSL_OSK_CANCEL) {
					char retbuff[INPUT_MAXLEN] = {0};
					oslOskGetText(retbuff);
					inbuff = retbuff;
				}
				oslEndOsk();
			};		
		
		// Check if there's a dialog open
		} else if (oslGetDialogType()) {
			if (oslGetDialogStatus() == PSP_UTILITY_DIALOG_NONE){
				oslEndDialog(); // Dialog closed
				
				switch (state) {
					case STATE_WIFI: {
						if (oslIsWlanConnected()) {
							SAY("** WiFi configuration completed.");
							SAY("** Please enter a nick to connect as.");
							inbuff = nick;
							state = STATE_NICK;
						} else {
							oslInitMessageDialog("Couldn't connect to WiFi!", 0);
							state = STATE_FATALWARNING;
						}
					} break;
					
					case STATE_FATALWARNING: {
						SAY("** Closed a fatal warning. Exiting...");
						runningFlag = false;
					} break;
					
					default: {
						SAY("** WARNING: Exited an unknown dialog (shouldn't happen!)");
					} break;
				}				
			}			
		
		// We're in the main screen
		} else {
			
			// Ensure wireless connectivity
			
			if (! oslIsWlanConnected()) {
				oslInitMessageDialog("Lost WiFi connection!", 0);
				state = STATE_FATALWARNING;
			}
			
			// Deal with ordinary keypresses
			
			if (osl_keys->released.circle) {
				// Bring up OSK
				oslInitOsk((char*)"", (char*)inbuff.c_str(), INPUT_MAXLEN-1, 2, -1 );
				
			} else if (osl_keys->released.square) {
				// Toggle debug messages
				if (debug) {
					SAY("** Debug messages DISABLED. Press \xFE to toggle.");
				} else {
					SAY("** Debug messages ENABLED. Press \xFE to toggle.");
				}
				debug = !debug;
				
			} else if (osl_keys->released.cross) {
				// Send message
				switch(state) {
					case STATE_NICK: {
						nick = inbuff;
						SAYF("** Nick set to \"%s\".", nick.c_str());
						
						SAY("** Please enter an address to connect to.");
						state = STATE_ADDRESS;
						inbuff = address;					
					} break;
					
					case STATE_ADDRESS: {
						address = inbuff;
						SAYF("** Connecting to \"%s\"...", address.c_str());
						
						state = STATE_TEXT;
						address = inbuff;
						inbuff = "";
						
						if (h) delete h;
						h = new NMDCHub(address.c_str(), nick.c_str(), 0);
					} break;
					
					case STATE_PASSWORD: {
						string pw = inbuff;
						SAY("** Logging in...");
						if (h) {
							h->Login(pw.c_str());
						}
						
						state = STATE_TEXT;
						inbuff = "";
					} break;
					
					case STATE_TEXT: {						
						if (h && h->Connected()) {
							h->SayPublic(inbuff.c_str());
						} else {
							SAY("** Not connected.");
						}
						inbuff = "";
					} break;
				
					default: {
						break;
					}
				};				
			
			} else if (osl_keys->released.triangle) {
				// Send help
				say("===============================================");
				say(" PSPDC Help");
				say("   [ ^ ] Display help");
				say("   [ o ] Enter text");
				say("   [ x ] Send message or enter command");
				say("   [ \xFE ] Toggle debug messages");
				say("===============================================");
				char ip[64] = {0}; oslGetIPaddress(ip);
				SAYF(" PSP IP address: %s", ip);
				say("===============================================");
			
			} // ...
			
		}
		
    }
	
    // Tidy up
	
	if (h) {
		h->Disconnect();
		delete h;
		h = NULL;
	}
	
    oslNetTerm();
    oslEndGfx();
    sceKernelExitGame();
    return 0;

}
